/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*                  COMPUTATION OF STARK UNITS                     */
/*                    OF TOTALLY REAL FIELDS                       */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/
/* $Id: stark.c,v 2.0.12.7 1998/11/06 15:07:57 belabas Exp belabas $ */
#include "pari.h"

/********************************************************************/
/*                    Miscellaneous functions                       */
/********************************************************************/

/* Compute the image of logelt by chi as a complex number if flag = 0,
   otherwise as a polmod, see InitChar in part 3 */
static GEN
ComputeImagebyChar(GEN chi, GEN logelt, long flag)
{
  if (!flag)
    return gpuigs((GEN)chi[2], itos(gmul((GEN)chi[1], logelt)));
  else
    return gpuigs((GEN)chi[4], itos(gmul((GEN)chi[1], logelt)));
}

/* Compute all the elements of a group given by its SNF */
static GEN
FindEltofGroup(long order, GEN cyc)
{
  long l, i, adec, j, dj;
  GEN rep, p1;

  l = lg(cyc)-1;

  rep = cgetg(order + 1, t_VEC);

  for  (i = 1; i <= order; i++)
  {
    p1 = cgetg(l + 1, t_COL);
    rep[i] = (long)p1;
    adec = i;

    for (j = l; j; j--)
    {
      dj = itos((GEN)cyc[j]);
      p1[j] = lstoi(adec%dj);
      adec /= dj;
    }
  }

  return rep;
}

/* Let dataC as given by InitQuotient0, compute a system of
   representatives of the quotient */
static GEN
ComputeLift(GEN dataC)
{
  long order, i, av = avma;
  GEN cyc, surj, eltq, elt;

  order = itos((GEN)dataC[1]);
  cyc   = (GEN)dataC[2];
  surj  = (GEN)dataC[3];

  eltq = FindEltofGroup(order, cyc);

  elt = cgetg(order + 1, t_VEC);

  for (i = 1; i <= order; i++)
    elt[i] = (long)inverseimage(surj, (GEN)eltq[i]);

  return gerepileupto(av, elt);
}

/* Let bnr1, bnr2 be such that mod(bnr2) | mod(bnr1), compute the
   matrix of the surjective map Cl(bnr1) ->> Cl(bnr2) */
static GEN
GetSurjMat(GEN bnr1, GEN bnr2)
{
  long nbg, i;
  GEN gen, M;

  gen = gmael(bnr1, 5, 3);
  nbg = lg(gen) - 1;

  M = cgetg(nbg + 1, t_MAT);
  for (i = 1; i <= nbg; i++)
    M[i] = (long)isprincipalray(bnr2, (GEN)gen[i]);

  return M;
}

/* Let chi a character defined over bnr and primitif over bnrc,
   compute the corresponding primitif character and the vectors of
   prime ideals dividing bnr but not bnr. */
static GEN
GetPrimChar(GEN chi, GEN bnr, GEN bnrc, long prec)
{
  long nbg, lc, i, j, l, av = avma, nd;
  GEN gen, cyc, U, d, rep, chic, M, s, p1, cond, condc, p2, nf, prdiff, Mrc, pi;

  gen   = gmael(bnr, 5, 3);
  nbg   = lg(gen) - 1;
  cyc   = gmael(bnr, 5, 2);
  Mrc   = diagonal(gmael(bnrc, 5, 2));
  nf    = gmael(bnr, 1, 7);
  cond  = gmael(bnr, 2, 1);
  condc = gmael(bnrc, 2, 1);
  pi    = mppi(prec);

  if (gegal(cond, condc))
    return NULL;

  cond  = (GEN)cond[1];
  condc = (GEN)condc[1];

  M  = GetSurjMat(bnr, bnrc);
  lc = lg((GEN)M[1]) - 1;
  p1 = hnfall(concat(M, Mrc));
  U  = (GEN)p1[2];

  chic = cgetg(lc + 1, t_VEC);
  for (i = 1; i <= lc; i++)
  {
    s  = gzero;
    p1 = (GEN)U[i + nbg];
    for (j = 1; j <= nbg; j++)
    {
      p2 = (GEN)cyc[j];
      p2 = gdiv((GEN)p1[j],p2);
      s  = gadd(s,gmul(p2,(GEN)chi[j]));
    }
    chic[i] = (long)s;
  }

  p2 = (GEN)idealfactor(nf, cond)[1];
  l  = lg(p2) - 1;
  nd = 0;

  prdiff = cgetg(l + 1, t_COL);
  for (i = 1; i <= l; i++)
    if (!idealval(nf, condc, (GEN)p2[i]))
    {
      nd++;
      prdiff[nd] = p2[i];
    }
  setlg(prdiff, nd + 1);

  p1  = cgetg(3, t_VEC);
  rep = cgetg(5, t_VEC);

  d = denom(chic);
  rep[1] = lmul(d, chic);
  rep[2] = lexp(gdiv(gmul2n(gmul(gi, pi), 1), d), prec);
  rep[3] = (long)d;
  rep[4] = lmodulcp(polx[0], cyclo(itos(d), 0));

  p1[1] = (long)rep;
  p1[2] = lcopy(prdiff);

  return gerepileupto(av,p1);
}

/* Let dataCR be a list of characters, compute the image of logelt */
static GEN
chiideal(GEN dataCR, GEN logelt, long flag)
{
  long j, cl;
  GEN rep;

  cl  = lg(dataCR) - 1;

  rep = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
    rep[j] = (long)ComputeImagebyChar(gmael(dataCR, j, 5), logelt, flag);

  return rep;
}

static long
GetMaxDeg(GEN dataCR)
{
  long i, l = lg(dataCR) - 1, s, mx = 0;

  for (i = 1; i <= l; i++)
  {
    s = phi(gmael3(dataCR, i, 5, 3))[2];
    if (s > mx) mx = s;
  }

  return mx;
}

/********************************************************************/
/*                    1rst part: find the field K                   */
/********************************************************************/

/* Let A be a finite abelian group given by its relation and let C
   define a subgroup of A, compute the order of A / C, its structure and
   the matrix expressing the generators of A on those of A / C */
static GEN
InitQuotient0(GEN DA, GEN C)
{
  long i, l;
  GEN MQ, MrC, H, snf, snf2, rep, p1;

  l = lg(DA)-1;
  H = gcmp0(C)? DA: C;
  MrC  = gauss(H, DA);
  snf  = smith2(hnf(MrC));
  MQ   = concatsp(gmul(H, (GEN)snf[1]), DA);
  snf2 = smith2(hnf(MQ));

  rep = cgetg(5, t_VEC);

  p1  = cgetg(l + 1, t_VEC);
  for (i = 1; i <= l; i++)
    p1[i] = lcopy(gcoeff((GEN)snf2[3], i, i));

  rep[1] = (long)dethnf((GEN)snf2[3]);
  rep[2] = (long)p1;
  rep[3] = lcopy((GEN)snf2[1]);
  rep[4] = lcopy(C);

  return rep;
}

/* Let m be a modulus et C a subgroup of Clk(m), compute all the data
   needed to work with the quotient Clk(m) / C namely 1) bnr(m), 2.1)
   its order, 2.2) its structure, 2.3) the matrix Clk(m) ->> Clk(m)/C
   and 2.4) the group C */
static GEN
InitQuotient(GEN bnr, GEN C)
{
  GEN Mrm, dataquo = cgetg(3, t_VEC);
  long av;

  dataquo[1] = lcopy(bnr);
  av = avma;  Mrm = diagonal(gmael(bnr, 5, 2));
  dataquo[2] = lpileupto(av, InitQuotient0(Mrm, C));

  return dataquo;
}

/* Let s: A -> B given by P, and let DA, DB be resp. the matrix of the
   relations of A and B, let nbA, nbB be resp. the rank of A and B,
   compute the kernel of s. If DA = 0 then A is free */
static GEN
ComputeKernel0(GEN P, GEN DA, GEN DB, long nbA, long nbB)
{
  long rk, av = avma;
  GEN herm, mask1, mask2, U;

  herm  = hnfall(concat(P, DB));
  rk = nbA + nbB + 1 - lg((GEN)herm[1]);

  mask1 = subis(shifti(gun, nbA), 1);
  mask2 = subis(shifti(gun, rk), 1);

  U = matextract((GEN)herm[2], mask1, mask2);

  if (gcmp0(DA))
    return gerepileupto(av, hnf(U));
  else
    return gerepileupto(av, hnf(concat(U, DA)));
}

/* Let m and n be two moduli such that n|m and let C be a congruence
   group modulo n, compute the corresponding congruence group modulo m
   ie then kernel of the map Clk(m) ->> Clk(n)/C */
static GEN
ComputeKernel(GEN bnrm, GEN dataC)
{
  long av = avma, i, nbm, nbq;
  GEN bnrn, Mrm, genm, Mrq, mgq, P;

  bnrn = (GEN)dataC[1];
  Mrm  = diagonal(gmael(bnrm, 5, 2));
  genm = gmael(bnrm, 5, 3);
  nbm  = lg(genm) - 1;
  Mrq  = diagonal(gmael(dataC, 2, 2));
  mgq  = gmael(dataC, 2, 3);
  nbq  = lg(mgq) - 1;

  P = cgetg(nbm + 1, t_MAT);
  for (i = 1; i <= nbm; i++)
    P[i] = lmul(mgq, isprincipalray(bnrn, (GEN)genm[i]));

  return gerepileupto(av, ComputeKernel0(P, Mrm, Mrq, nbm, nbq));
}

/* Let C a congruence group in bnr, compute its subgroups of index as
   subgroups of Clk(bnr) */
static GEN
ComputeIndex2Subgroup(GEN bnr, GEN C)
{
  long nbsbgrp, i, l, av = avma;
  GEN Mrc, snf, Mr, U, subgrp, rep, p1;

  Mr = diagonal(gmael(bnr, 5, 2));

  Mrc = gmul(ginv(C), Mr);
  snf = smith2(Mrc);

  U = ginv((GEN)snf[1]);
  l = lg((GEN)snf[3]);

  p1 = cgetg(l, t_VEC);
  for (i = 1; i < l; i++)
    p1[i] = mael3(snf, 3, i, i);

  subgrp  = subgrouplist0(p1, 2, 0, 0);
  nbsbgrp = lg(subgrp) - 1;

  rep = cgetg(nbsbgrp, t_VEC);
  for (i = 2; i <= nbsbgrp; i++)
  {
    p1 = hnf(concat(gmul(C, gmul(U, (GEN)subgrp[i])), diagonal(Mr)));
    rep[i - 1] = (long)p1;
  }

  return gerepileupto(av, rep);
}

/* Let pr be a prime (pr may divide mod(bnr)), compute the indexes
   e,f of the splitting of pr in the class field nf(bnr/subgroup) */
static GEN
GetIndex(GEN pr, GEN bnr, GEN subgroup, long prec)
{
  long av = avma, v, lg, i;
  GEN bnf, mod, mod0, mpr0, mpr, bnrpr, subpr, M, e, f, dtQ, p1, rep, cycpr, cycQ;

  bnf  = (GEN)bnr[1];
  mod  = gmael(bnr, 2, 1);
  mod0 = (GEN)mod[1];

  /* Compute the part of mod coprime to pr */
  v = idealval(bnf, mod0, pr);
  mpr0 = idealdivexact(bnf, mod0, idealpow(bnf, pr, stoi(v)));

  mpr = cgetg(3, t_VEC);
  mpr[1] = (long)mpr0;
  mpr[2] = mod[2];
  if (gegal(mpr0, mod0))
  {
    bnrpr = bnr;
    subpr = subgroup;
  }
  else
  {
    bnrpr = buchrayinitgen(bnf, mpr, prec);
    cycpr = gmael(bnrpr, 5, 2);
    M = GetSurjMat(bnr, bnrpr);
    M = gmul(M, subgroup);
    subpr = hnf(concat(M, diagonal(cycpr)));
  }

  /* e = #(bnr/subgroup) / #(bnrpr/subpr) */
  e = gdiv(det(subgroup), det(subpr));

  /* f = order of [pr] in bnrpr/subpr */
  dtQ  = InitQuotient(bnrpr, subpr);
  p1   = isprincipalray(bnrpr, pr);
  p1   = gmul(gmael(dtQ, 2, 3), p1);
  cycQ = gmael(dtQ, 2, 2);
  lg = lg(cycQ) - 1;
  f  = gun;
  for (i = 1; i <= lg; i++)
    f = glcm(f, gdiv((GEN)cycQ[i], ggcd((GEN)cycQ[i], (GEN)p1[i])));

  rep = cgetg(3, t_VEC);
  rep[1] = lcopy(e);
  rep[2] = lcopy(f);

  return gerepileupto(av, rep);
}

/* Let f be a conductor without infinite part and let C be a
   congruence group modulo f, compute (m,D) such that D is a
   congruence group of conductor m where m is a multiple of f
   divisible by all the infinite places but one, D is a subgroup of
   index 2 of Im(C) in Clk(m), no prime dividing f splits in the
   corresponding quadratic extension and m is of minimal norm. Return
   bnr(m), D, quotient Ck(m) / D and Clk(m) / C */
static GEN
FindModulus(GEN dataC, long prec)
{
  long n, i, narch, nbp, maxnorm, minnorm, flag, N, nbidnn, s, c;
  long j, nbcand, av = avma, av1, limnorm, debkp;
  GEN bnr, rep, bnf, nf, f, vecun, arch, m, listid, idnormn, bnrm, ImC;
  GEN  candD, D, bpr, indpr, sgp, p1;

  bnr = (GEN)dataC[1];
  sgp = gmael(dataC, 2, 4);
  bnf = (GEN)bnr[1];
  nf  = (GEN)bnf[7];
  N   = degree((GEN)nf[1]);
  f   = gmael3(bnr, 2, 1, 1);

  debkp = DEBUGLEVEL;
  DEBUGLEVEL = 0;

  bpr = (GEN)idealfactor(nf, f)[1];
  nbp = lg(bpr) - 1;

  indpr = cgetg(nbp + 1,t_VEC);
  for (i = 1; i <= nbp; i++)
  {
    p1 = GetIndex((GEN)bpr[i], bnr, sgp, prec);
    indpr[i] = lmulii((GEN)p1[1], (GEN)p1[2]);
  }

  /* Initialization of the possible infinite part */
  vecun = cgetg(N + 1, t_VEC);
  for (i = 1; i <= N; i++) vecun[i] = un;

  if (N == 2) narch = 1; /* if N=2, only one case is necessary */
  else narch = N;

  arch = cgetg(narch + 1, t_VEC);
  for (i = 1; i <= narch; i++)
  {
    arch[i] = lcopy(vecun);
    mael(arch, i, N + 1 - i) = zero;
  }

  m = cgetg(3, t_VEC);

  /* we go from minnorm up to maxnorm, if necessary we increase this values.
     If we cannot find a suitable conductor of norm < limnorm, we stop */
  maxnorm = 50;
  minnorm = 1;
  limnorm = 200;

  if (debkp >= 2)
    fprintferr("Looking for a modulus of norm: ");

  flag = 1;
  while (flag)
  {

    /* compute all ideals of norm <= maxnorm */
    listid = ideallist(nf, maxnorm);

    av1 = avma;
    for (n = minnorm; (n <= maxnorm) && flag; n++)
    {
      if (debkp >= 2) fprintferr(" %ld", n);

      idnormn = (GEN)listid[n];
      nbidnn  = lg(idnormn) - 1;
      for (i = 1; (i <= nbidnn) && flag; i++)
      {
	avma = av1;

        /* finite part of the conductor */
	m[1] = (long)idealmul(nf, f, (GEN)idnormn[i]);

	for (s = 1; ( s<= narch) && flag; s++)
	{
	  /* infinite part */
	  m[2] = arch[s];

          /* compute Clk(m), check if m is a conductor */
	  bnrm = buchrayinitgen(bnf, m, prec);
	  if (gcmp1(conductor(bnr, gzero, -1, prec)))
	  {
	    /* compute Im(C) in Clk(m)... */
	    ImC = ComputeKernel(bnrm, dataC);

	    /* ... and its subgroups of index 2 */
	    candD  = ComputeIndex2Subgroup(bnrm, ImC);
	    nbcand = lg(candD) - 1;
	    for (c = 1; ( c <= nbcand) && flag; c++)
	    {
	      /* check if m is the conductor */
	      D = (GEN)candD[c];
	      flag = !itos(conductor(bnrm, D, -1, prec));

	      if (!flag)
	      {
		/* check the splitting of primes */
		for (j = 1; (j <= nbp) && !flag; j++)
		{
		  p1 = GetIndex((GEN)bpr[j], bnrm, D, prec);
		  p1 = mulii((GEN)p1[1], (GEN)p1[2]);

		  if (gegal(p1, (GEN)indpr[j])) flag = 1;
		}
	      }
	    }
	  }
	}
      }
    }
    if (flag)
    {
      /* if necessary compute more ideals */
      minnorm = maxnorm;
      maxnorm <<= 1;

      if (maxnorm > limnorm)
	err(talker, "Cannot find a suitable modulus in FindModulus");
    }
  }

  if (debkp >= 2) fprintferr("\n");

  DEBUGLEVEL = debkp;

  rep = cgetg(5, t_VEC);
  rep[1] = lcopy(bnrm);
  rep[2] = lcopy(D);
  rep[3] = (long)InitQuotient((GEN)rep[1], (GEN)rep[2]);
  rep[4] = (long)InitQuotient((GEN)rep[1], ImC);

  return gerepileupto(av, rep);
}

/********************************************************************/
/*                       2d part: compute W(X)                      */
/********************************************************************/

/* compute W(chi) such that Ld(s,chi) = W(chi) Ld(1 - s, chi*),
   if flag != 0 do not check the result */
static GEN
ComputeArtinNumber(GEN datachi, long flag, long prec)
{
  long av = avma, av2, G, ms, j, i, nz, zcard, q, l, N;
  GEN chi, nc, dc, p1, cond0, cond1, elts, Msign, umod2, lambda, nf, sg, p2;
  GEN chib, diff, vt, z, idg, mu, idh, zid, zstruc, zgen, zchi, allclass, class;
  GEN bnr, beta, s, tr, p3, den, muslambda, pi;

  chi   = (GEN)datachi[8];
  bnr   = (GEN)datachi[3];
  nf    = gmael(bnr, 1, 7);
  diff  = gmael(nf, 5, 5);
  cond0 = gmael3(bnr, 2, 1, 1);
  cond1 = gmael3(bnr, 2, 1, 2);
  umod2 = gmodulcp(gun, gdeux);
  N     = degree((GEN)nf[1]);
  pi    = mppi(prec);

  /* trivial case */
  if (cmpsi(2, (GEN)chi[3]) >= 0) return gerepileupto(av, gun);

  /* some initializations */
  G   = - bit_accuracy(prec) >> 1;
  nc  = idealnorm(nf, cond0);
  dc  = idealmul(nf, diff, cond0);
  den = idealnorm(nf, dc);
  z   = gexp(gdiv(gmul2n(gmul(gi, pi), 1), den), prec);

  q = 0;
  for (i = 1; i < lg(cond1); i++)
    if (gcmp1((GEN)cond1[i])) q++;

  /* compute a system of elements congru to 1 mod cond0 and giving all
     possible signatures for cond1 */
  p1 = zarchstar(nf, cond0, cond1, q);
  elts = (GEN)p1[2];
  Msign = gmul((GEN)p1[3], umod2);
  ms = lg(elts) - 1;

  /* find lambda in diff.cond such that gcd(lambda.(diff.cond)^-1,cond0) = 1 and
     lambda >(cond1)> 0 */
  lambda = idealappr(nf, dc);
  sg = zsigne(nf, lambda, cond1);
  p2 = lift(gmul(Msign, sg));

  for (j = 1; j <= ms; j++)
    if (gcmp1((GEN)p2[j])) lambda = element_mul(nf, lambda, (GEN)elts[j]);

  idg = idealdivexact(nf, lambda, dc);

  /* find mu in idg such that idh=(mu) / idg is coprime with cond0 and
     mu >(cond1)> 0 */
  if (!gcmp1(gcoeff(idg, 1, 1)))
  {
    p1 = idealfactor(nf, idg);
    p2 = idealfactor(nf, cond0);

    l = lg((GEN)p2[1]) - 1;
    for (i = 1; i <= l; i++)
      coeff(p2, i, 2) = zero;

    p1 = gtrans(concat(gtrans(p1), gtrans(p2)));
    mu = idealapprfact(nf, p1);
    sg = zsigne(nf, mu, cond1);
    p2 = lift(gmul(Msign, sg));

    for (j = 1; j <= ms; j++)
      if (gcmp1((GEN)p2[j])) mu = element_mul(nf, mu, (GEN)elts[j]);

    idh = idealdivexact(nf, mu, idg);
  }
  else
  {
    mu  = gun;
    idh = gcopy(idg);
  }

  muslambda = element_div(nf, mu, lambda);

  /* compute a system of generators of (Ok/cond)^* cond1-positive */
  zid = zidealstarinitgen(nf, cond0);

  zcard  = itos(gmael(zid, 2, 1));
  zstruc = gmael(zid, 2, 2);
  zgen   = gmael(zid, 2, 3);
  nz = lg(zgen) - 1;

  zchi = cgetg(nz + 1, t_VEC);
  for (i = 1; i <= nz; i++)
  {
    p1 = (GEN)zgen[i];
    sg = zsigne(nf, p1, cond1);
    p2 = lift(gmul(Msign, sg));

    for (j = 1; j <= ms; j++)
      if (gcmp1((GEN)p2[j])) p1 = element_mul(nf, p1, (GEN)elts[j]);

    class = isprincipalray(bnr, p1);
    zchi[i] = (long)ComputeImagebyChar(chi, class, 0);
    zgen[i] = (long)p1;
  }

  /* Sum chi(beta) * exp(2i * pi * Tr(beta * mu / lambda) where beta
     runns through the classes of (Ok/cond0)^* and beta cond1-positive */
  allclass = FindEltofGroup(zcard, zstruc);

  p3 = cgetg(N + 1, t_COL);
  for (i = 1; i <= N; i++)
    p3[i] = zero;

  vt = cgetg(N + 1, t_VEC);
  for (i = 1; i <= N; i++)
  {
    p3[i] = un;
    vt[i] = ltrace(basistoalg(nf, p3));
    p3[i] = zero;
  }

  s = cgetg(3, t_COMPLEX);
  s[1] = lgetr(prec);
  s[2] = lgetr(prec);
  gaffect(gzero, s);

  av2 = avma;
  for (i = 1; i <= zcard; i++)
  {
    beta = gun;
    chib = gun;
    p1 = (GEN)allclass[i];

    for (j = 1; j <= nz; j++)
    {
      p2 = element_powmodideal(nf, (GEN)zgen[j], (GEN)p1[j], cond0);
      beta = element_mul(nf, beta, p2);
      chib = gmul(chib, gpuigs((GEN)zchi[j], itos((GEN)p1[j])));
    }

    sg = zsigne(nf, beta, cond1);
    p2 = lift(gmul(Msign, sg));

    for (j = 1; j <= ms; j++)
      if (gcmp1((GEN)p2[j]))
	beta = element_mul(nf, beta, (GEN)elts[j]);

    beta = element_mul(nf, beta, muslambda);
    tr = gmul(vt, beta);
    tr = gmod(gmul(tr, den), den);

    gaffect(gadd(s, gmul(chib, gpuigs(z, itos(tr)))), s);

    avma = av2;
  }

  class = isprincipalray(bnr, idh);
  s = gmul(s, ComputeImagebyChar(chi, class, 0));
  s = gdiv(s, gsqrt(nc, prec));

  p1 = gsubgs(gabs(s, prec), 1);

  i = expo(p1);
  if ((i > G) && !flag)
    err(bugparier, "ComputeArtinNumber");

  return gerepileupto(av, gmul(s, gpuigs(gneg_i(gi),q)));
}

/* compute the constant W of the functional equation of
   Lambda(chi). If flag = 1 then chi is assumed to be primitive */
GEN
bnrrootnumber(GEN bnr, GEN chi, long flag, long prec)
{
  long av = avma, l, i;
  GEN cond, condc, bnrc, chic, cyc, d, p1, p2, dtcr, pi;

  checkbnr(bnr);

  cond = gmael(bnr, 2, 1);
  l    = lg(gmael(bnr, 5, 2));
  pi   = mppi(prec);

  if ((typ(chi) != t_VEC) || (lg(chi) != l))
    err(talker, "incorrect character in bnrrootnumber");

  if (!flag)
  {
    condc = bnrconductorofchar(bnr, chi, prec);
    if (gegal(cond, condc)) flag = 1;
  }
  else condc = cond;

  if (flag)
    bnrc = bnr;
  else
    bnrc = buchrayinitgen((GEN)bnr[1], condc, prec);

  chic = cgetg(l, t_VEC);
  cyc  = gmael(bnr, 5, 2);
  for (i = 1; i < l; i++)
    chic[i] = ldiv((GEN)chi[i], (GEN)cyc[i]);
  d = denom(chic);

  p2 = cgetg(4, t_VEC);
  p2[1] = lmul(d, chic);
  p2[2] = lexp(gdiv(gmul2n(gmul(gi, pi), 1), d), prec);
  if (gegal(d, gdeux))
    p2[2] = lreal((GEN)p2[2]);
  p2[3] = (long)d;

  dtcr = cgetg(9, t_VEC);
  dtcr[1] = (long)chi;
  dtcr[2] = zero;
  dtcr[3] = (long)bnrc;
  dtcr[4] = (long)bnr;
  dtcr[5] = (long)p2;
  dtcr[6] = zero;
  dtcr[7] = (long)condc;

  p1 = GetPrimChar(chi, bnr, bnrc, prec);

  if (!p1)
    dtcr[8] = (long)p2;
  else
    dtcr[8] = p1[1];

  return gerepileupto(av, ComputeArtinNumber(dtcr, 0, prec));
}


/********************************************************************/
/*               3d part: initialize the characters                 */
/********************************************************************/

static GEN
LiftChar(GEN cyc, GEN Mat, GEN chi)
{
  long lm, l, i, j, av;
  GEN lchi, s;

  lm = lg(cyc) - 1;
  l  = lg(chi) - 1;

  lchi = cgetg(lm + 1, t_VEC);
  for (i = 1; i <= lm; i++)
  {
    av = avma;
    s  = gzero;

    for (j = 1; j <= l; j++)
      s = gadd(s, gmul((GEN)chi[j], gcoeff(Mat, j, i)));

    lchi[i] = (long)gerepileupto(av, gmod(gmul(s, (GEN)cyc[i]), (GEN)cyc[i]));
  }

  return lchi;
}

/* Let chi be a character, compute A(chi) corresponding to the primes
   dividing diff */
static GEN
ComputeAChi(GEN dtcr)
{
  long l, i;
  GEN ray, A, diff, chi, bnrc;

  diff = (GEN)dtcr[6];
  bnrc = (GEN)dtcr[3];
  chi  = (GEN)dtcr[8];
  l    = lg(diff) - 1;

  A = gun;
  for (i = 1; i <= l; i++)
  {
    ray = isprincipalray(bnrc, (GEN)diff[i]);
    A   = gmul(A, gsub(gun, ComputeImagebyChar(chi, ray, 0)));
  }

  return A;
}

/* Remark: a character is given by a vector [(c_i), z, d, pm] such that
   chi(id) = z^s where s = sum(c_i * a_i) and a_i is the log of id on
   the generators of bnr, z is exp(2i * pi / d) and pm is z as a polmod */
/* compute a vector dataCR contening for each character: 1) the usual
   representation of chi in bnr(m), 2) C(chi), 3) bnr(cond(chi)), 4)
   bnr(m), 5) [(c_i), z, d, pm] in bnr(m), 6) diff(chi) primes dividing m
   but not cond(chi), 7) finite part of cond(chi) and 8)
   [(c_i), z, d, pm] in bnr(cond(chi)) */
static GEN
InitChar(GEN dataD, long prec)
{
  long nbg, hD, h, lD, i, nc, tnc, j, N, av = avma, fl, tetpil, prec2;
  GEN bnr, Mr, mod0, bnf, MrD, Surj, p1, dataCR, chi, lchi, cond, bnrc;
  GEN chic, dk, C, d, p2, p3, setchar, toto, pi;

  bnr   = (GEN)dataD[1];
  Mr    = gmael(bnr, 5, 2);
  mod0  = gmael3(bnr, 2, 1, 1);
  nbg   = lg(Mr) - 1;
  bnf   = (GEN)bnr[1];
  hD    = itos(gmael(dataD, 2, 1));
  h     = hD >> 1;
  MrD   = gmael(dataD, 2, 2);
  Surj  = gmael(dataD, 2, 3);
  lD    = lg(MrD)-1;
  dk    = gmael(bnf, 7, 3);
  N     = degree(gmael(bnf, 7, 1));
  prec2 = prec<<1;
  pi    = mppi(prec2);

  setchar = gtoset(cgetg(1, t_VEC));

  toto = cgetg(3, t_VEC);
  toto[1] = toto[2] = zero;

  p1 = FindEltofGroup(hD, MrD);

  dataCR = cgetg(h + 1, t_VEC);
  for (i = 1; i <= h ;i++) dataCR[i] = lgetg(9, t_VEC);

  C = gsqrt(gdiv(dk, gpuigs(pi, N)), prec2);
  nc  = 1;
  /* tnc is the number of characters computed so far */
  tnc = 1;


  for (i = 1; tnc <= h; i++)
  {
    /* lift a character of D in Clk(m) */
    chi = gcopy((GEN)p1[i]);

    for (j = 1; j <= lD; j++) chi[j] = ldiv((GEN)chi[j], (GEN)MrD[j]);
    lchi = LiftChar(Mr, Surj, chi);

    /* find its conductor */
    if (setsearch(setchar, lchi, 0)) cond = toto;
    else
      cond = bnrconductorofchar(bnr, lchi, prec);

    /* if the infinite part of chi is non trivial then chi is ok */
    if (!gcmp0((GEN)cond[2]))
    {
      /* does we know already about the invariants of chi? */
      fl = 0;
      for (j = 1; (j < nc) && !fl; j++)
	if (gegal((GEN)cond[1], gmael(dataCR, j, 7))) fl = j;

      if (!fl && gegal((GEN)cond[1], mod0))
      {
	mael(dataCR, nc, 2) = lmul(C, gsqrt(det((GEN)cond[1]), prec2));
	mael(dataCR, nc, 3) = lcopy(bnr);
	mael(dataCR, nc, 4) = (long)idmat(nbg);
	mael(dataCR, nc, 6) = lgetg(1, t_VEC);
	mael(dataCR, nc, 7) = lcopy(mod0);
	
	fl = nc;
      }

      /* constant C(chi) */
      mael(dataCR, nc, 1) = (long)lchi;
      if (!fl)
	mael(dataCR, nc, 2) = lmul(C, gsqrt(det((GEN)cond[1]), prec2));
      else
	mael(dataCR, nc, 2) = mael(dataCR, fl, 2);

      /* bnr(cond(chi)) */
      if (!fl)
     	bnrc = buchrayinitgen(bnf, cond, prec);
      else
	bnrc = gmael(dataCR, fl, 3);

      mael(dataCR, nc, 3) = (long)bnrc;

      /* bnr(m) */
      mael(dataCR, nc, 4) = (long)bnr;

      /* compute [(c_i), z, d] corresponding to bnr(m) */
      chic = cgetg(nbg+1, t_VEC);
      for (j = 1; j <= nbg; j++)
	chic[j] = ldiv((GEN)lchi[j], (GEN)Mr[j]);
      d = denom(chic);

      p2 = cgetg(5, t_VEC);
      p2[1] = lmul(d, chic);
      p2[2] = lexp(gdiv(gmul2n(gmul(gi, pi), 1), d), prec);
      p2[3] = (long)d;
      p2[4] = lmodulcp(polx[0], cyclo(itos(d), 0));
      mael(dataCR, nc, 5) = (long)p2;

      /* compute diff(chi) and the corresponding primitive character */
      mael(dataCR, nc, 7) = cond[1];
      p3 = GetPrimChar(lchi, bnr, bnrc, prec);
      if (p3)
      {	
	mael(dataCR, nc, 6) = p3[2];
	mael(dataCR, nc, 8) = p3[1];
      }
      else
      {
	mael(dataCR, nc, 6) = lgetg(1, t_VEC);
	mael(dataCR, nc, 8) = (long)p2;
      }
	
      /* If chi is not real,  add its conjugate character */
      if (!gegal(d, gdeux))
      {
	p3 = cgetg(nbg + 1, t_VEC);

	for (j = 1; j <= nbg; j++)
	  p3[j] = lsub((GEN)Mr[j], (GEN)lchi[j]);

	p2 = cgetg(2, t_VEC);
	p2[1] = (long)p3;
	
	setchar = setunion(setchar, p2);
	tnc++;
      }

      nc ++;
      tnc++;

      p2 = cgetg(2, t_VEC);
      p2[1] = (long)lchi;
      setchar = setunion(setchar, p2);
    }
  }

  setlg(dataCR, nc);

  tetpil = avma;
  return gerepile(av, tetpil, gcopy(dataCR));
}

/* recompute dataCR with the new precision */
static GEN
CharNewPrec(GEN dataCR, GEN nf, long prec)
{
  GEN dk, C, p1, pi;
  long av = avma, N, l, j, prec2;

  dk    =  (GEN)nf[3];
  N     =  degree((GEN)nf[1]);
  l     =  lg(dataCR) - 1;
  prec2 = prec << 1;
  pi    = mppi(prec2);

  C = gsqrt(gdiv(dk, gpuigs(pi, N)), prec2);

  for (j = 1; j <= l; j++)
  {
    mael(dataCR, j, 2) = lmul(C, gsqrt(det(gmael(dataCR, j, 7)), prec2));

    mael4(dataCR, j, 3, 1, 7) = lcopy(nf);

    p1 = gmael(dataCR, j, 5);
    p1[2] = lexp(gdiv(gmul2n(gmul(gi, pi), 1), (GEN)p1[3]), prec);

    p1 = gmael(dataCR, j, 8);
    p1[2] = lexp(gdiv(gmul2n(gmul(gi, pi), 1), (GEN)p1[3]), prec);
  }

  return gerepileupto(av, gcopy(dataCR));
}

/********************************************************************/
/*             4th part: compute the coefficients an(chi)           */
/********************************************************************/

/* Remark: the coefficients an are computed as arrays of longs. An
   array contains the coefficients of an(chi) as a polmod modulo
   cyclo(order(chi)) */

/* initialize the matrix matan */
static int***
InitMatAn(long cl, long nmax, long mx, long flag)
{
  long i, j, k;
  int ***matan;

  matan = (int***)gpmalloc((cl + 1) * sizeof(int**));

  for (i = 0; i <= cl; i++)
  {
    matan[i] = (int**)gpmalloc((nmax + 1) * sizeof(int*));

    for (j = 0; j <= nmax; j++)
    {
      matan[i][j] = (int*)gpmalloc(mx * sizeof(int));

      for (k = 0; k < mx; k++)
	if (((j == 1) || flag) && !k) matan[i][j][k] = 1;
	else matan[i][j][k] = 0;
    }
  }

  return matan;
}

/* free the n*m*l array */
static void
FreeArray(int ***a, long n, long m)
{
  long i, j;

  for (i = 0; i <= n; i++)
  {
    for (j = 0; j <= m; j++)
      free(a[i][j]);

    free(a[i]);
  }
  free(a);

  return;
}

/* transform a polmod into coeff */
static int*
Polmod2Coeff(GEN polmod, long mx)
{
  GEN pol2;
  long i;
  int *rep;

  pol2 = (GEN)polmod[2];

  rep = (int*)gpmalloc(mx * sizeof(int));

  for (i = 0; i < mx; i++)
    rep[i] = itos(truecoeff(pol2, i));

  return rep;
}

/* initialize the reduction of the coeff */
static int***
InitReduction(GEN dataCR, long mx)
{
  GEN d, x, polmod, pol;
  long dg, i, j, cl;
  int ***reduc;

  cl = lg(dataCR) - 1;

  reduc = (int***)gpmalloc((cl + 1) * sizeof(int**));

  for (i = 0; i <= cl; i++)
  {
    reduc[i] = (int**)gpmalloc(mx * sizeof(int*));

    if (i)
    {
      d   =  gmael3(dataCR, i, 5, 3);
      pol =  cyclo(itos(d), 0);
      x   =  polx[0];
      dg  =  degree(pol);
    }

    for (j = 0; j < mx; j++)
      if (i)
      {
	polmod = gmodulcp(gpuigs(x, dg + j), pol);
	reduc[i][j] = Polmod2Coeff(polmod, mx);
      }
      else
	reduc[i][j] = (int*)gpmalloc(mx * sizeof(int));
  }

  return reduc;
}

/* add two coeff */
static int*
AddCoeff(int* c1, int* c2, long mx)
{
  long i;
  int *c3;

  c3 = (int*)gpmalloc(mx * sizeof(int));

  for (i = 0; i < mx; i++)
    c3[i] = c1[i] + c2[i];

  return c3;
}

/* multiply (with reduction) a polmod with a coeff */
static int*
MulPolmodCoeff(GEN polmod, int* c1, int** reduc, long mx)
{
  long dg, i, j;
  int *c2, *c3, *c4;

  if(gcmp1(polmod)) return c1;
  dg  =  degree((GEN)polmod[1]);

  c2 = Polmod2Coeff(polmod, mx);
  c3 = (int*)gpmalloc(2 * dg * sizeof(int));

  for (i = 0; i<= 2 * dg - 1; i++)
  {
    c3[i] = 0;

    for (j = 0; j <= i; j++)
      if ((j < mx) && (j > i - mx))
	c3[i] = c3[i] + c1[j] * c2[i-j];
  }

  c4 = (int*)gpmalloc(mx * sizeof(int));

  for (i = 0; i < dg; i++)
  {
    c4[i] = c3[i];
    for (j = 0; j < dg; j++)
      c4[i] = c4[i] + reduc[j][i] * c3[dg+j];
  }
  for (i = dg; i < mx; i++) c4[i] = 0;

  free(c2);
  free(c3);

  return c4;
}

/* evaluate the coeff. No Garbage collector*/
static GEN
EvalCoeff(GEN z, int* c1, long dg)
{
  long i;
  GEN e;

  e = stoi(c1[dg - 1]);
  for (i = dg - 2; i >= 0; i--)
    e = gadd(stoi(c1[i]), gmul(z, e));

  return e;
}

/* copy the n * m array matan */
static void
CopyCoeff(int*** array, int*** array2, long n, long m, long mx)
{
  long i, j, k;

   for (i = 0; i <= n; i++)
   {
     for (j = 0; j <= m; j++)
     {
       for (k = 0; k < mx; k++)
	 array2[i][j][k] = array[i][j][k];
     }
   }

   return;
}

/* initialize the data for GetRay */
static GEN
InitGetRay(GEN bnr,  long nmax)
{
  long bd, i, j, l, dbk;
  GEN listid, listcl, id, rep, bnf, cond;

  bnf  =  (GEN)bnr[1];
  cond =  gmael3(bnr, 2, 1, 1);

  if (nmax < 1000) return NULL;
  else bd = min(1000, nmax / 50);

  rep = cgetg(3, t_VEC);

  dbk = DEBUGLEVEL;
  DEBUGLEVEL = 0;
  listid = ideallist(bnf, bd);
  DEBUGLEVEL = dbk;

  listcl = cgetg(bd + 1, t_VEC);
  for (i = 1; i <= bd; i++)
  {
    l = lg((GEN)listid[i]) - 1;
    listcl[i] = lgetg(l + 1, t_VEC);

    for (j = 1; j <= l; j++)
    {
      id = gmael(listid, i, j);
      if (gcmp1(gcoeff(idealadd(bnf, id, cond), 1, 1)))
	mael(listcl, i, j) = (long)isprincipalray(bnr, id);
    }
  }

  if (DEBUGLEVEL) msgtimer("InitGetRay");

  rep[1] = (long)listid;
  rep[2] = (long)listcl;

  return rep;
}

/* compute the class of the prime ideal pr in cl(bnr) using dataray */
static GEN
GetRay(GEN bnr,  GEN dataray,  GEN pr)
{
  long av = avma, N, n, bd, c;
  GEN id, tid, t2, u, alpha, p1, cl, listid, listcl, nf, cond;

  if (!dataray)
    return isprincipalray(bnr, pr);

  listid =  (GEN)dataray[1];
  listcl =  (GEN)dataray[2];
  cond   =  gmael3(bnr, 2, 1, 1);
  bd     =  lg(listid) - 1;
  nf     =  gmael(bnr, 1, 7);
  t2     =  gmael(nf, 5, 4);
  N      =  degree((GEN)nf[1]);

  id = idealhermite(nf, pr);
  tid = qf_base_change(t2, id);
  u = lllgramint(tid);

  c = 1; alpha = NULL;
  while (!alpha && (c <= N))
  {
    p1 = gmul(id, (GEN)u[c++]);
    if (gcmp1(gcoeff(idealadd(nf, p1, cond), 1, 1)))
      alpha = p1;
  }
  if (!alpha)
    return gerepileupto(av, isprincipalray(bnr, pr));

  id = idealdivexact(nf, alpha, id);

  n = itos(det(id));
  if (n > bd)
    cl = isprincipalray(bnr, id);
  else
  {
    cl = NULL;
    c  = 1;
    p1 = (GEN)listid[n];
    while (!cl)
    {
      if (gegal((GEN)p1[c], id))
	cl = gmael(listcl, n, c);
      c++;
    }
  }

  return gerepileupto(av, gsub(isprincipalray(bnr, alpha), cl));
}

/* correct the coefficients an(chi) according with diff(chi) */
static int**
CorrectCoeff(GEN dtcr, int** an, int** reduc, long nmax, long mx)
{
  long lg, av1, j, p, q, limk, k, l, av = avma;
  int ***an2, *c, *d;
  GEN chi, bnrc, diff, ray, ki, ki2, pr;

  chi  =  (GEN)dtcr[8];
  bnrc =  (GEN)dtcr[3];
  diff =  (GEN)dtcr[6];
  lg   =  lg(diff) - 1;

  if (DEBUGLEVEL > 2)
  {
    fprintferr("diff(chi) = ");
    output(diff);
  }

  an2 = InitMatAn(1, nmax, mx, 0);
  av1 = avma;
  if (!lg)
    return an;
  else
  {
    for(j = 1; j <= lg; j++)
    {
      for (k = 0; k <= nmax; k++)
	for (l = 0; l < mx; l++)
	  an2[1][k][l] = an[k][l];

      pr  = (GEN)diff[j];
      ray = isprincipalray(bnrc, pr);
      ki  = ComputeImagebyChar(chi, ray, 1);
      ki2 = gcopy(ki);

      p = itos(gpui((GEN)pr[1], (GEN)pr[4], 0));
      q = p;
      limk = nmax / q;

      while (q <= nmax)
      {
	for(k = 1; k <= limk; k++)
	{
	  d = MulPolmodCoeff(ki2, an2[1][k], reduc, mx);
	  c = AddCoeff(an[k * q], d, mx);
	  an[k * q] = c;
	}

	q *= p; limk /= p;
	ki2 = gmul(ki2, ki);
      }
      avma = av1;
    }
  }

  FreeArray(an2, 1, nmax);

  avma = av;
  return an;
}

/* compute the coefficients an in the general case */
static int***
ComputeCoeff(GEN dataCR, long nmax)
{
  long cl, i, j, av = avma, av2, np, q, limk, k, id, cpt = 200, mx;
  int *c,*d, ***matan, ***reduc, ***matan2;
  GEN tabprem, bnf, pr, cond, ray, ki, ki2, prime, npg, bnr, dataray;
  byteptr dp = diffptr;

  bnr  =  gmael(dataCR, 1, 4);
  bnf  =  (GEN)bnr[1];
  cond =  gmael3(bnr, 2, 1, 1);
  cl   =  lg(dataCR) - 1;

  dataray = InitGetRay(bnr, nmax);

  mx = GetMaxDeg(dataCR);
  matan  = InitMatAn(cl, nmax, mx, 0);
  matan2 = InitMatAn(cl, nmax, mx, 0);
  reduc  = InitReduction(dataCR, mx);

  prime = stoi(2);
  dp++;
  av2 = avma;
  while (*dp && (prime[2] <= nmax))
  {
    tabprem = primedec(bnf, prime);

    for (j = 1; j < lg(tabprem); j++)
    {
      pr  = (GEN)tabprem[j];
      np  = VERYBIGINT;
      npg = gpui((GEN)pr[1], (GEN)pr[4], 0);
      if (cmpis(npg, VERYBIGINT) < 0) np = itos(npg);

      if (!idealval(bnf, cond, pr) && (np <= nmax))
      {
	CopyCoeff(matan, matan2, cl, nmax, mx);

	ray = GetRay(bnr, dataray, pr);
	ki  = chiideal(dataCR, ray, 1);
	ki2 = gcopy(ki);

	q = np;
	limk = nmax / q;

	while (q <= nmax)
	{
	  for (k = 1; k <= limk; k++)
	    for (id = 1; id <= cl; id++)
	    {
	      d = MulPolmodCoeff((GEN)ki2[id], matan2[id][k], reduc[id], mx);
	      c = AddCoeff(matan[id][k * q], d, mx);

	      matan[id][k * q] = c;
	    }

	  q *= np;
	  limk /= np;
	
	  for (id = 1; id <= cl; id++)
	    ki2[id] = lmul((GEN)ki2[id], (GEN)ki[id]);
	}
      }
    }

    prime[2] += (*dp++);
    avma = av2;

    if ((DEBUGLEVEL > 1) && (prime[2] > cpt))
    {
      fprintferr(" %ld ", prime[2]);
      cpt += 200;
    }
  }

  if (DEBUGLEVEL > 1) fprintferr("\n");

  for (i = 1; i <= cl; i++)
    matan[i] = CorrectCoeff((GEN)dataCR[i], matan[i], reduc[i], nmax, mx);

  FreeArray(matan2, cl, nmax);
  FreeArray(reduc, cl, mx - 1);

  avma = av;
  return matan;
}

/********************************************************************/
/*              5th part: compute L functions at s=1                */
/********************************************************************/

/* This part should be re-written someday... */

static long
tmpGetBoundN0(GEN C,  long N,  long prec)
{
  long av = avma, N0;
  GEN gN, mu, alpha, beta, eps, A0, c1, c0, c2, lneps, limx, racpi, pi;

  pi = mppi(prec);

  gN = stoi(N);
  mu = gmul2n(gN, -1);
  racpi = gsqrt(pi, prec);

  alpha = gmul2n(stoi(N + 3), -1);
  beta  = gpui(gdeux, gmul2n(gN, -1), 3);

  eps = gdiv(gpuigs(stoi(10), - (prec - 3) / pariK1), gdeux);

  A0 = gmul2n(gun, N);
  A0 = gmul(A0, gpuigs(mu, N + 2));
  A0 = gmul(A0, gpuigs(gmul2n(pi, 1), 1 - N));
  A0 = gsqrt(A0, 3);

  c1 = gmul(mu, gpui(beta, ginv(mu), 3));
  c0 = gdiv(gmul(A0, gpuigs(gmul(gdeux, pi), N - 1)), mu);
  c0 = gmul(c0, gpui(c1, gsub(gun, alpha), 3));
  c2 = gdiv(gsub(alpha, gun), mu);

  lneps = glog(gdiv(c0, eps), 3);
  limx = gdiv(gsub(glog(lneps, 3), glog(c1, 3)), gadd(c2, gdiv(lneps, mu)));
  limx = gmul(gpui(gdiv(c1, lneps), mu, 3), gadd(gun, gmul(c2, gmul(mu, limx))));

  N0 = itos(gfloor(gdiv(C, limx)));
  avma = av;

  return N0;
}

static long
tmpGetBoundi0(long N,  long prec)
{
  long av = avma, imin, i0, itest;
  GEN gN, mu, alpha, beta, eps, A0, c1, c0, c2, lneps, limx, borneps, ftest, racpi, pi;

  pi = mppi(prec);

  gN = stoi(N);
  mu = gmul2n(gN, -1);
  racpi = gsqrt(pi, prec);

  alpha = gmul2n(stoi(N + 3), -1);
  beta  = gpui(gdeux, gmul2n(gN, -1), 3);

  eps = gdiv(gpuigs(stoi(10), - (prec - 3) / pariK1), gdeux);

  A0 = gmul2n(gun, N);
  A0 = gmul(A0, gpuigs(mu, N + 2));
  A0 = gmul(A0, gpuigs(gmul2n(pi, 1), 1 - N));
  A0 = gsqrt(A0, 3);

  c1 = gmul(mu, gpui(beta, ginv(mu), 3));
  c0 = gdiv(gmul(A0, gpuigs(gmul(gdeux, pi), N - 1)), mu);
  c0 = gmul(c0, gpui(c1, gsub(gun, alpha), 3));
  c2 = gdiv(gsub(alpha, gun), mu);

  lneps = glog(gdiv(c0, eps), 3);
  limx  = gdiv(gsub(glog(lneps, 3), glog(c1, 3)), gadd(c2, gdiv(lneps, mu)));
  limx  = gmul(gpui(gdiv(c1, lneps), mu, 3), gadd(gun, gmul(c2, gmul(mu, limx))));

  borneps = gpuigs(stoi(5), N);
  borneps = gdiv(borneps, gmul(gmul(gsqrt(limx, 3), eps), gpuigs(racpi, 3)));

  imin = 1;
  i0   = 1400;
  while(i0 - imin >= 4)
  {
    itest = (i0 + imin) >> 1;

    ftest = gpuigs(limx, itest);
    ftest = gmul(ftest, gpuigs(mpfactr(itest / 2, 3), N));

    if(gcmp(ftest, borneps) >= 0)
      i0 = itest;
    else
      imin = itest;
  }

  i0 = (i0 / 2) * 2;

  avma = av;
  return i0;
}

static GEN
tmpGetST(GEN dataCR,  long prec)
{
  GEN zet, c_pair, ck_pair, c_impair, ck_impair, serie_pair, serie_impair;
  GEN serie_exp, aij, racpi, S, T, rep, csurn, ipk, tab, lncsurn, p1, nsurc;
  GEN lncsurnj, p12, an, p2, p21, B, D, bnr, bnf, C, *gptr[3], cond, pi, eul;
  long i, j, k, n, av = avma, av2, tetpil, N, hk, fj, id, prec2, i0, nmax;
  long test, N0[100], CC[100], mx;
  int ***matan;

  bnr   = gmael(dataCR, 1, 4);
  bnf   = (GEN)bnr[1];
  N     = degree(gmael(bnf, 7, 1));
  hk    = lg(dataCR) - 1;
  prec2 = prec << 1;
  pi    = mppi(prec2);
  eul   = mpeuler(prec2);

  racpi = gsqrt(pi, prec2);

  C    = cgetg(hk + 1, t_VEC);
  cond = cgetg(hk + 1, t_VEC);
  for (i = 1; i <= hk; i++)
  {
    C[i]    = mael(dataCR, i, 2);
    cond[i] = mael(dataCR, i, 7);
  }

  i0 = tmpGetBoundi0(N, prec);
  nmax = 0;
  for (i = 1; i <= hk; i++)
  {
    N0[i] = tmpGetBoundN0((GEN)C[i], N, prec);

    /* FIXME: we don't want N0[i] to be too big */
    N0[i] = min(N0[i], 20000);

    nmax = max(nmax, N0[i]);
  }

  if(DEBUGLEVEL >= 2) fprintferr("nmax = %ld\n", nmax);

  matan = ComputeCoeff(dataCR, nmax);
  mx = GetMaxDeg(dataCR);
  if (DEBUGLEVEL) msgtimer("Compute an");

  av2 = avma;
  zet = cgetg(N + 2, t_VEC);
  zet[1] = (long)eul;
  for (i = 2; i <= N+1; i++)
    zet[i] = (long)gzeta(stoi(i), prec2);

  aij = cgetg(i0+1, t_VEC);
  for (i = 1; i <= i0; i++)
    aij[i] = lgetg(N + 2, t_VEC);

  c_pair    = gmul2n(gpuigs(racpi, N - 1), 1);
  c_impair  = gneg_i(gmul2n(racpi, N));
  ck_pair   = cgetg(3, t_VEC);
  ck_impair = cgetg(N + 1, t_VEC);

  for(k = 1; k <= 2; k++)
  {
      ck_pair[k] = lmul((GEN)zet[k], gaddsg(N - 1, gmul2n(stoi(2 - N), - k)));
      if (k & 1) ck_pair[k] = lneg((GEN)ck_pair[k]);
  }

  for (k = 1; k <= N; k++)
  {
    ck_impair[k] = lmul((GEN)zet[k], gadd(gun, gmul2n(stoi(N - 2), - k)));
    if (k & 1) ck_impair[k] = lneg((GEN)ck_impair[k]);
    ck_impair[k] = ladd((GEN)ck_impair[k], gun);
  }

  ck_pair[1]   = lsub((GEN)ck_pair[1], gmulsg(N - 1, glog(gdeux, prec2)));
  ck_impair[1] = lsub((GEN)ck_impair[1], glog(gdeux, prec2));

  serie_pair = cgetg(4, t_SER);
  setsigne(serie_pair, 1);
  setvarn(serie_pair, 0);
  setvalp(serie_pair, 1);

  serie_impair=cgetg(N + 2, t_SER);
  setsigne(serie_impair, 1);
  setvarn(serie_impair, 0);
  setvalp(serie_impair, 1);

  for (k = 1; k <= 2; k++)
    serie_pair[k + 1] = ldiv((GEN)ck_pair[k], stoi(k));
  serie_exp = gmul(c_pair, gexp(serie_pair, prec2));

  for (j = 1; j <= 2; j++) mael(aij, 1, j) = serie_exp[4 - j];
  for (k = 1; k <= N; k++) serie_impair[k + 1]=ldiv((GEN)ck_impair[k], stoi(k));

  serie_exp = gmul(c_impair, gexp(serie_impair, prec2));

  for (j = 1; j <= N; j++) mael(aij, 2, j) = serie_exp[N + 2 - j];
  for (j = 3; j <= N + 1; j++) mael(aij, 1, j) = zero;
  mael(aij, 2, N + 1) = zero;

  affsr(1, c_pair = cgetr(prec2));
  c_pair = gmul2n(c_pair, N);

  c_impair = gneg_i(gdiv(gmul(gmulgs(c_pair, 4), racpi), stoi(6)));
  c_pair   = gmul(c_pair, gpuigs(racpi, N - 1));
  if (N & 1)
  {
    c_pair   = gneg_i(c_pair);
    c_impair = gneg_i(c_impair);
  }

  for (k = 1; k <= 2; k++)
  {
    ck_pair[k] = lmul((GEN)zet[k], gaddsg(N - 1, gmul2n(stoi(2 - N), - k)));
    if (k & 1) ck_pair[k] = lneg((GEN)ck_pair[k]);
    ck_pair[k] = ladd((GEN)ck_pair[k], gadd(stoi(N - 1), gpuigs(gdeux, - k)));
  }

  for (k = 1; k <= N; k++)
  {
    ck_impair[k] = lmul((GEN)zet[k], gadd(gun, gmul2n(stoi(N - 2), - k)));
    if (k & 1) ck_impair[k] = lneg((GEN)ck_impair[k]);
    ck_impair[k] = ladd(gmul2n(stoi(N - 2), - k), (GEN)ck_impair[k]);
    ck_impair[k] = ladd((GEN)ck_impair[k],
			gadd(gun, gadd(gpuigs(gdeux, - k), gpuigs(stoi(3), - k))));
  }

  ck_pair[1]   = lsub((GEN)ck_pair[1], gmul(stoi(N - 1), glog(gdeux, prec2)));
  ck_impair[1] = lsub((GEN)ck_impair[1], glog(gdeux, prec2));

  i = 1;
  while (i < i0 / 2)
  {
    for (k = 1; k <= 2; k++)
      serie_pair[k + 1] = ldiv((GEN)ck_pair[k], stoi(k));
    serie_exp = gmul(c_pair, gexp(serie_pair, prec2));

    for (j = 1; j <= 2; j++)
      mael(aij, 2 * i + 1, j) = serie_exp[4 - j];

    for (k = 1; k <= N; k++)
      serie_impair[k + 1] = ldiv((GEN)ck_impair[k], stoi(k));

    serie_exp = gmul(c_impair, gexp(serie_impair, prec2));

    for (j = 1; j <= N; j++)
	mael(aij, 2 * i + 2, j)=serie_exp[N + 2 - j];

    for (j = 3; j <= N + 1; j++)
      mael(aij, 2 * i + 1, j) = zero;
    mael(aij, 2 * i + 2, N + 1) = zero;

    i++;
    c_pair   = gmul2n(gmul(c_pair, gpuigs(stoi(i - 1), N - 1)), 2 * N - 2);
    c_pair   = gdiv(c_pair, gmulsg(i, gpuigs(stoi(4 * i * i - 6 * i + 2), N - 1)));
    c_impair = gmul(c_impair, gdivsg(4, gmul(gpuigs(stoi(i), N - 2),
					     stoi(4 * i * i + 2 * i))));

    if (N & 1)
    {
      c_pair   = gneg_i(c_pair);
      c_impair = gneg_i(c_impair);
    }

    for (k = 1; k <= 2; k++)
      ck_pair[k] = ladd((GEN)ck_pair[k], gadd(ginv(gpuigs(stoi(2 * i), k)),
					      gdivsg(N - 1, gpuigs(stoi(2 * i - 1), k))));

    for (k = 1; k <= N; k++)
      ck_impair[k] = ladd((GEN)ck_impair[k],
			  gadd(gdivsg(N - 1, gpuigs(stoi(2 * i), k)),
			       ginv(gpuigs(stoi(2 * i + 1), k))));
  }

  tetpil = avma;
  aij = gerepile(av2, tetpil, gcopy(aij));
  if (DEBUGLEVEL) msgtimer("Compute a(ij)");

  av2 = avma;

  tab = cgetg(N + 1, t_VEC);
  for (k = 0; k <= N - 1; k++)
  {
    tab[k + 1] = lgetg(N + 3, t_MAT);

    for (j = 1; j <= N + 2; j++)
    {
      mael(tab, k + 1, j)=lgetg(i0 + 1, t_COL);
      if (j < 3) fj = 1;
      else fj *= (j - 2);

      for (i = 1; i <= i0; i++)
	if (j > N - k)
	  coeff((GEN)tab[k + 1], i, j) = zero;
	else
	  coeff((GEN)tab[k + 1], i, j) = ldivgs((GEN)((GEN)aij[i])[j + k], fj);
    }
  }

  ipk = cgetg(i0 + 1, t_MAT);
  for (i = 1; i <= i0; i++)
  {
    ipk[i] = lgetg(N + 1, t_COL);

    p1 = gdivgs(gun, i);
    for (k = 0; k <= N - 1; k++)
    {
      coeff(ipk, k + 1, i) = lcopy(p1);
      p1 = gdivgs(p1, i);
    }
  }

  B = cgetg(N + 3, t_MAT);
  D = cgetg(N + 3, t_MAT);
  for (j = 1; j <= N + 2; j++)
  {
    B[j] = lgetg(i0 + 1, t_COL);
    D[j] = lgetg(i0 + 1, t_COL);
    for (i = 1; i <= i0; i++)
    {
      coeff(B, i, j) = zero;
      coeff(D, i, j) = zero;

      for (k = 0; k <= N - 1; k++)
      {
	coeff(B, i, j) = ladd((GEN)coeff(B, i, j),
			      gmul((GEN)coeff((GEN)tab[k + 1], i, j),
				   (GEN)coeff(ipk, k + 1, i)));
	if (i > 1)
	  coeff(D, i, j) = ladd((GEN)coeff(D, i, j),
				gmul((GEN)coeff((GEN)tab[k + 1], i, j),
				     (GEN)coeff(ipk, k + 1, i - 1)));
      }
    }
  }

  gptr[0] = &B;
  gptr[1] = &D;
  gptr[2] = &tab;
  gerepilemany(av2, gptr, 3);

  if (DEBUGLEVEL) msgtimer("Initialize B&D");

  p1 = cgetg(3, t_COMPLEX);
  p1[1] = lgetr(prec);
  p1[2] = lgetr(prec);
  gaffect(gzero, p1);

  S = cgetg(hk + 1, t_VEC);
  T = cgetg(hk + 1, t_VEC);
  for (id = 1; id <= hk; id++)
  {
    S[id] = lcopy(p1);
    T[id] = lcopy(p1);

    test = 0;
    for (k = 1; (k < id) && !test; k++)
      if (gegal((GEN)cond[id], (GEN)cond[k]))
	test = k;

    if (test)
      CC[id] = test;
    else
      CC[id] = id;
  }

  av2 = avma;

  for (id = 1; id <= hk; id++)
  {
    if (CC[id] == id)
    {
      for (n = 1; n <= N0[id]; n++)
      {
	csurn    = gdivgs((GEN)C[id], n);
	lncsurn  = glog(csurn, prec2);
	lncsurnj = gun;

	p1 = gmul(csurn, racpi);
	p2 = gzero;
	
	for (j = 1; j <= N + 1; j++)
	{
	  p21 = (GEN)coeff(D, i0, j + 1);
	  p12 = (GEN)coeff(B, i0, j +1 );
	  nsurc = ginv(csurn);
	
	  for (i = i0; i > 1; i--)
	  {
	    p12 = gadd(gmul(p12, nsurc), (GEN)coeff(B, i - 1, j + 1));
	    p21 = gadd(gmul(p21, nsurc), (GEN)coeff(D, i - 1, j + 1));
	  }
	
	  p21 = gsub((GEN)coeff((GEN)tab[1], 1, j), p21);
	  p1  = gsub(p1, gmul(p12, lncsurnj));
	  p2  = gadd(p2, gmul(p21, lncsurnj));

	  lncsurnj = gmul(lncsurnj, lncsurn);
	}

	for (j = 1; j <= hk; j++)
	  if (CC[j] == id)
	  {
	    an = EvalCoeff(gmael3(dataCR, j, 5, 2), matan[j][n], mx);

	    gaffect(gadd((GEN)S[j], gmul(p1, an)), (GEN)S[j]);
	    gaffect(gadd((GEN)T[j], gmul(p2, gconj(an))), (GEN)T[j]);
	  }

	avma = av2;

	if (!(n%200) && (DEBUGLEVEL > 1)) fprintferr(" %ld ", n);
      }
    }
  }

  FreeArray(matan, hk, nmax);

  if (DEBUGLEVEL)
  {
    fprintferr("\n");
    msgtimer("Compute S&T");
  }

  racpi = gmul2n(gsqrt(gpuigs(pi, N - 1), prec), 1);

  rep = cgetg(3, t_VEC);
  rep[1] = lconj(gdiv(S, racpi));
  rep[2] = lconj(gdiv(T, racpi));

  return gerepileupto(av, gcopy(rep));
}


/********************************************************************/
/*                6th part: recover the coefficients                */
/********************************************************************/

static long
TestOne(GEN plg,  GEN beta,  GEN B,  long v,  long G,  long N)
{
  long av = avma, ok, j;
  GEN p1;

  p1 = gsub(beta, (GEN)plg[v]);

  if (expo(p1) < G)
  {
    ok = 1;
    for (j = 1; (j <= N) && ok; j++)
      if (j != v)
      {
	p1 = gabs((GEN)plg[j], DEFAULTPREC);
	if (gcmp(p1, B) > 0) ok = 0;
      }
  }
  else
  {
    avma = av;
    return 0;
  }

  avma = av;
  return ok;
}

/* to be re-written someday... */
static GEN
SearchPoint(GEN a,  long b,  GEN beta,  GEN B,  GEN M,  GEN gam,
            long nv,  long G,  long N,  long prec)
{
  GEN Q, u, r, unr, cand, plg, eta, eta2;
  long n1 = lg(a), n, nonnul, ok = 0, i, j, k, s, *x, av = avma, av2;
  double p, **q, *v, *y, *z, eps = 0.000001;

  if (DEBUGLEVEL > 1) fprintferr("In SearchPoint with b = %ld and G = %ld\n", b, G);

  n = n1 - 1;
  affsr(1, unr = cgetr(10));

  x =   (long*)gpmalloc(n1 * sizeof(long));
  y = (double*)gpmalloc(n1 * sizeof(double));
  z = (double*)gpmalloc(n1 * sizeof(double));
  v = (double*)gpmalloc(n1 * sizeof(double));

  q = (double**)gpmalloc(n1 * sizeof(double*));
  for (j = 1;j <= n; j++)
    q[j] = (double*)gpmalloc(n1 * sizeof(double));

  u = lllgramint(a);
  Q = qf_base_change(a, u);
  Q = gmul(Q, unr);
  r = sqred1(Q);

  for (j = 1;j <= n; j++)
  {
    v[j] = rtodbl(gcoeff(r, j, j));

    for (i = 1;i < j; i++)
      q[i][j] = rtodbl(gcoeff(r, i, j));
  }

  eta2 = cgetg(n1, 18);
  for (i = 1;i <= n;i++) eta2[i] = lgeti(10);

  av2 = avma;

  s = 1;
  k = n;
  y[n] = z[n] = 0;
  x[n] = (long)(sqrt(b / v[n] + eps));

  do
  {
    do
    {
      if (k > 1)
      {
	k--;
	z[k] = 0;
	for (j = k + 1; j <= n; j++)
	  z[k] = z[k] + q[k][j] * x[j];

	p = x[k + 1] + z[k + 1];

	y[k] = y[k + 1] + p * p * v[k + 1];
	x[k] = (long)floor(sqrt((b - y[k] + eps) / v[k]) - z[k]);
      }

      while (v[k] * (x[k] + z[k]) * (x[k] + z[k]) > b - y[k] + eps)
      {
	k++;
	x[k]--;
      }

    }
    while((k > 1) && !ok);

    if ((nonnul = (x[1] || (y[1] > eps))))
    {
      for (i = 1; i <= n; i++)
	affsi(x[i], (GEN)eta2[i]);
      eta = gmul(u, eta2);

      cand = gadd(eta, gam);
      plg  = gmul(M, cand);

      ok = TestOne(plg, beta, B, nv, G, N);

      if (!ok)
      {
	cand = gsub(gam, eta);
	plg  = gmul(M, cand);
	ok   = TestOne(plg, beta, B, nv, G, N);
      }

      if (ok)
	cand = gerepileupto(av2, gcopy(cand));
      else avma = av2;
 	
      x[k]--;
    }
  }
  while(nonnul && !ok);

  free(x);
  free(y);
  free(z);
  free(v);
  for (j = 1; j <= n; j++)
    free(q[j]);
  free(q);

  if (ok)
    return gerepileupto(av, gcopy(cand));
  else
    return NULL;
}

/* Using enumeration of lattice points */
static GEN
RecCoeff1(GEN nf,  GEN beta,  GEN B,  long v,  long prec)
{
  long i, N, G, av = avma, b, maxb = 100000;
  GEN vec, gam, M, T2, plg, nB, rep;

  M  = gmael(nf, 5, 1);
  T2 = gmael(nf, 5, 4);
  N  = degree((GEN)nf[1]);

  G = min(-20, - bit_accuracy(prec) >> 4);

  vec = cgetg(N + 1, t_COL);
  for (i = 1; i <= N; i++)
    if ( i== v)
      vec[i] = (long)beta;
    else
      vec[i] = zero;

  gam = ground(inverseimage(M, vec));
  plg = gmul(M, gam);

  if (TestOne(plg, beta, B, v, G, N))
    return gerepileupto(av, basistoalg(nf, gam));

  nB = gsqr(gsub((GEN)plg[v], beta));
  for (i = 1; i <= N; i++)
    if (i!=v) nB = gadd(nB, gsqr((GEN)plg[i]));

  nB = gceil(gadd(nB, mulsi(N, gsqr(B))));

  /* FIXME: the bound 10 * maxb is rather arbitrary */
  if (cmpis(nB, 10 * maxb) < 0)
    b  = itos(nB);
  else
    b = 10 * maxb;

  rep = NULL;

  rep = SearchPoint(T2, min(b, maxb), beta, B, M, gam, v, G, N, prec);

  if (!rep && (b > maxb))
    rep = SearchPoint(T2, b, beta, B, M, gam, v, G, N, prec);

  if (rep) return gerepileupto(av, basistoalg(nf, rep));
  avma = av;
  return NULL;
}

/* Using linear dependance relations */
static GEN
RecCoeff2(GEN nf,  GEN beta,  GEN B,  long v,  long prec)
{
  long N, G, ok, i, bacmin, bacmax, av = avma, av2;
  GEN vec, velt, p1, cand, M, plg, pol, cand2;

  M    = gmael(nf, 5, 1);
  pol  = (GEN)nf[1];
  N    = degree(pol);
  vec  = gtrans((GEN)gtrans(M)[v]);
  velt = (GEN)nf[7];

  G = min( - 20, - bit_accuracy(prec) >> 4);

  p1 = cgetg(2, t_VEC);

  p1[1] = lneg(beta);
  vec = concat(p1, vec);

  p1[1] = zero;
  velt = concat(p1, velt);

  bacmin = (long)(.225 * bit_accuracy(prec));
  bacmax = (long)(.315 * bit_accuracy(prec));

  av2 = avma;

  ok = 0;
  for (i = bacmax; (i >= bacmin) && !ok; i--)
  {
    p1 = lindep2(vec, i);

    if (signe((GEN)p1[1]))
    {
      p1    = ground(gdiv(p1, (GEN)p1[1]));
      cand  = gmodulcp(gmul(velt, gtrans(p1)), pol);
      cand2 = algtobasis(nf, cand);
      plg   = gmul(M, cand2);

      ok = TestOne(plg, beta, B, v, G, N);
    }

    if (!ok) avma = av2;
  }

  if (ok) return gerepileupto(av, gcopy(cand));

  avma = av;
  return NULL;
}

static GEN recbeta2(GEN nf,  GEN beta,  GEN bound,  long prec);

/* Attemps to find an algebraic integer close to beta at the place v
   and less than B at all the others */
GEN
RecCoeff(GEN nf,  GEN polnum,  long v,  long flag, long prec)
{
  long av = avma, cl, j;
  GEN p1, pol, bound, beta, Bmax = stoi(200000);

  cl = degree(polnum);

  pol = gcopy(polnum);
  for (j = 2; j <= cl + 1; j++)
  {
    bound = binome(stoi(cl), j - 2);
    if (flag < 3) bound = shifti(bound, cl + 2 - j);

    beta = greal((GEN)polnum[j]);

    if (DEBUGLEVEL > 1)
      fprintferr("In RecCoeff with B = %Z\n", (long)bound);

    p1 = RecCoeff2(nf, beta, bound, v, prec);

    if (!p1)
    {
      if (gcmp(bound, Bmax) > 0)
	if (degree((GEN)nf[1]) == 2)
	  p1 = recbeta2(nf, beta, Bmax, prec);
	else
	  p1 = RecCoeff1(nf, beta, Bmax, v, prec);
    }

    if (!p1)
    {	
      if (degree((GEN)nf[1]) == 2)
	p1 = recbeta2(nf, beta, bound, prec);
      else
	p1 = RecCoeff1(nf, beta, bound, v, prec);
    }

    if (!p1) return NULL;

    pol[j] = (long)p1;
  }

  return gerepileupto(av, gcopy(pol));
}


/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*                   Computation of class fields of                */
/*	          real quadratic fields using Stark units          */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/


/* FindModulus for quadhilbertreal,  return Clk(m),  the group D,  the
   quotient Cl(m)/D,  dataCR of Cl(m)/D and the quotient Cl(m)/C */
GEN
FindModulus0(GEN bnf,  long prec)
{
  GEN bnr, dataC, p1, rep;

  bnr   = buchrayinitgen(bnf, gun, prec);
  dataC = InitQuotient(bnr, gzero);

  rep = cgetg(6, t_VEC);
  p1 = FindModulus(dataC, prec);
  rep[1] = p1[1];
  rep[2] = p1[2];
  rep[3] = p1[3];
  rep[5] = p1[4];
  rep[4] = (long)InitChar((GEN)rep[3], prec);

  return rep;
}

/* compute the coefficients an for the quadratic case */
static int***
computean(GEN dtcr,  long nmax)
{
  long i, j, cl, q, cp, al, v1, v2, v, fldiv, av, av1, mx;
  int ***matan, ***reduc;
  GEN bnf, ideal, dk, idno, p1, prime, chi, qg, chi1, chi2, zpol;
  GEN chi11, chi12, bnr, pr, pr1, pr2, xray, xray1, xray2, dataray;
  byteptr dp = diffptr;

  cl = lg(dtcr) - 1;
  mx = GetMaxDeg(dtcr);

  matan = InitMatAn(cl, nmax, mx, 1);
  reduc = InitReduction(dtcr, mx);

  zpol = cgetg(3, t_VEC);

  p1 = gpuigs(polx[0], mx - 1);
  for (i = 0; i < mx; i++)
    p1[i + 2] = zero;
  zpol[1] = un;
  zpol[2] = (long)p1;

  av = avma;

  bnr = gmael(dtcr, 1, 4);
  bnf = (GEN)bnr[1];
  dataray = InitGetRay(bnr, nmax);

  ideal = gmael3(bnr, 2, 1, 1);
  idno  = idealnorm(bnf, ideal);
  dk = gmael(bnf, 7, 3);

  prime = stoi(2);
  dp++;

  av1 = avma;

  while (*dp && (prime[2] <= nmax))
  {
    qg = prime;
    al = 1;

    switch (krogs(dk, prime[2]))
    {
      /* prime is inert */
      case -1:
	fldiv = divise(idno, prime);

	if (!fldiv)
	{
	  xray = GetRay(bnr, dataray, prime);
	  chi  = chiideal(dtcr, xray, 1);
	  chi1 = gcopy(chi);
	}

       	while(cmpis(qg, nmax) <= 0)
	{
	  q = qg[2];

	  for (cp = 1, i = q; i <= nmax; i += q, cp++)
	    if(cp % prime[2])
	    {
	      if (fldiv || (al % 2))
                for (j = 1; j <= cl; j++)
		  matan[j][i] = Polmod2Coeff(zpol, mx);
	      else
		for (j = 1; j <= cl; j++)
		  matan[j][i] = MulPolmodCoeff((GEN)chi[j], matan[j][i], reduc[j], mx);
	    }

	  qg = mulii(qg, prime);
	  al++;
	
	  if ((al % 2) && !fldiv)
	    for (j = 1; j <= cl; j++)
	      chi[j] = lmul((GEN)chi[j], (GEN)chi1[j]);
 	}
	break;

    /* prime is ramified */
    case 0:
      fldiv = divise(idno, prime);

      if (!fldiv)
      {
	pr   = (GEN)primedec(bnf, prime)[1];
	xray = GetRay(bnr, dataray, pr);
	chi  = chiideal(dtcr, xray, 1);
	chi2 = gcopy(chi);
      }

      while(cmpis(qg, nmax) <= 0)
      {
	q = qg[2];
	
	for (cp = 1, i = q; i <= nmax; i += q, cp++)
	  if(cp % prime[2])
          {
	    if (fldiv)
	      for(j = 1; j <= cl; j++)
		matan[j][i] = Polmod2Coeff(zpol, mx);
	    else
            {
	      for(j = 1; j <= cl; j++)
		matan[j][i] = MulPolmodCoeff((GEN)chi[j], matan[j][i], reduc[j], mx);
	    }
	  }
	
	qg = mulii(qg, prime);
	al++;
	
	if((cmpis(qg, nmax) <= 0) && !fldiv)
	  for (j = 1; j <= cl; j++)
	    chi[j] = lmul((GEN)chi[j], (GEN)chi2[j]);
      }
      break;

    /* prime is split */
    case 1:
      p1  = primedec(bnf, prime);
      pr1 = (GEN)p1[1];
      pr2 = (GEN)p1[2];
      v1 = idealval(bnf, ideal, pr1);
      v2 = idealval(bnf, ideal, pr2);

      if (v1 + v2 == 0)
      {
	xray1 = GetRay(bnr, dataray, pr1);
	xray2 = GetRay(bnr, dataray, pr2);
	chi11 = chiideal(dtcr, xray1, 1);
	chi12 = chiideal(dtcr, xray2, 1);

	chi1 = gadd(chi11, chi12);
	chi2 = gcopy(chi12);

	while(cmpis(qg, nmax) <= 0)
        {
	  q = qg[2];
	  for (cp = 1, i = q; i <= nmax; i += q, cp++)
	    if(cp % prime[2])
	      for(j = 1; j <= cl; j++)
		matan[j][i] = MulPolmodCoeff((GEN)chi1[j], matan[j][i], reduc[j], mx);

	  qg = mulii(qg, prime);
	  al++;

	  if(cmpis(qg, nmax) <= 0)
	    for (j = 1; j <= cl; j++)
            {
	      chi2[j] = lmul((GEN)chi2[j], (GEN)chi12[j]);
	      chi1[j] = ladd((GEN)chi2[j], gmul((GEN)chi1[j], (GEN)chi11[j]));
	    }
	}
      }
      else
      {
	if (v1)
        {
	  v  = v2;
	  pr = pr2;
	}
	else
        {
	  v  = v1;
	  pr = pr1;
	}
	
	if (v == 0)
        {
	  xray = GetRay(bnr, dataray, pr);
	  chi1 = chiideal(dtcr, xray, 1);
	  chi  = gcopy(chi1);
	}
	
	while(cmpis(qg, nmax) <= 0)
        {
	  q = qg[2];
	  for (cp = 1, i = q; i <= nmax; i += q, cp++)
	    if(cp % prime[2])
            {
	      if (v)
		for (j = 1; j <= cl; j++)
		  matan[j][i]=Polmod2Coeff(zpol, mx);
	      else
		for (j = 1; j <= cl; j++)
		  matan[j][i]=MulPolmodCoeff((GEN)chi[j], matan[j][i], reduc[j], mx);
	    }
	
	  qg = mulii(qg, prime);
	  al++;

	  if (!v && (cmpis(qg, nmax) <= 0))
	    for (j = 1; j <= cl; j++)
	      chi[j] = lmul((GEN)chi[j], (GEN)chi1[j]);
	}
      }
      break;
    }

    prime[2] += (*dp++);

    avma = av1;
  }

  for (i = 1; i <= cl; i++)
    matan[i]=CorrectCoeff((GEN)dtcr[i], matan[i], reduc[i], nmax, mx);

  FreeArray(reduc, cl, mx - 1);

  avma = av;
  return matan;
}


/* compute S and T for the quadratic case */
static GEN
QuadGetST(GEN data,  long prec)
{
  long av = avma, n, j, nn[100], nmax, cl, h, av1, av2, test, k, flag, mx;
  int ***matan;
  GEN C, p1, c2, cexp, cn, cn2, veclprime2, veclprime1, subgroup, dtcr, cond, rep, an;

  subgroup = (GEN)data[2];
  dtcr     = (GEN)data[4];
  flag     = 0;
  cl       = lg(dtcr) - 1;
  h        = itos(gmul2n(det(subgroup), -1));
  mx       = GetMaxDeg(dtcr);

  constpi(prec + 1);

  C    = cgetg(cl+1, t_VEC);
  cond = cgetg(cl+1, t_VEC);
  nmax = 0;
  for (j = 1; j <= cl; j++)
  {
    C[j]    = mael(dtcr, j, 2);
    cond[j] = mael(dtcr, j, 7);
    nn[j]   = (bit_accuracy(prec) * gtodouble((GEN)C[j]) * 0.35);

    /* FIXME: we don't want nn[j] to be too big */
    nn[j] = min(nn[j], 10000);

    nmax = max(nmax, nn[j]);
  }

  if (DEBUGLEVEL >= 2)
    fprintferr("nmax = %ld\n", nmax);

  /* compute the coefficients */
  matan = computean(dtcr, nmax);
  if (DEBUGLEVEL) msgtimer("Compute an");

  /*  allocate memory for the answer */
  rep = cgetg(3, t_VEC);

  /* allocate memory for veclprime1 */
  veclprime1 = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
  {
    p1 = cgetg(3, t_COMPLEX);
    p1[1] = lgetr(prec);
    p1[2] = lgetr(prec);
    gaffect(gzero, p1);

    veclprime1[j] = (long)p1;
  }

  /* compute f2(C/n) */
  av1 = avma;

  c2   = cgetg(cl + 1, t_VEC);
  cexp = cgetg(cl + 1, t_VEC);
  cn   = cgetg(cl + 1, t_VEC);
  cn2  = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
  {
    c2[j]   = ldivsg(2, (GEN)C[j]);
    cexp[j] = lexp(gneg_i((GEN)c2[j]), prec);
    cn[j]   = lgetr(prec); affsr(1, (GEN)cn[j]);
    cn2[j]  = lgetr(prec);
  }

  av2 = avma;

  /* compute veclprime1 */
  for (j = 1; j <= cl; j++)
  {
    for (n = 1; n <= nn[j]; n++)
    {
      affrr(gmul((GEN)cn[j], (GEN)cexp[j]), (GEN)cn[j]);
      affrr(gdivgs((GEN)cn[j], n), (GEN)cn2[j]);

      an = EvalCoeff(gmael3(dtcr, j, 5, 2), matan[j][n], mx);

      gaffect(gadd((GEN)veclprime1[j], gmul((GEN)cn2[j], an)), (GEN)veclprime1[j]);

      avma = av2;
    }
  }

  for (j = 1; j <= cl; j++)
    gaffect(gmul2n(gconj(gmul((GEN)veclprime1[j], (GEN)C[j])), - 1), (GEN)veclprime1[j]);

  avma = av1;
  rep[1] = (long)veclprime1;

  if (DEBUGLEVEL) msgtimer("Compute V1");

  /* allocate memory for veclprime1 */
  veclprime2 = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
  {
    p1 = cgetg(3, t_COMPLEX);
    p1[1] = lgetr(prec);
    p1[2] = lgetr(prec);
    gaffect(gzero, p1);

    veclprime2[j] = (long)p1;
  }

  /* compute f1(C/n) */
  av1 = avma;

  c2 = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
    c2[j] = ldivsg(2, (GEN)C[j]);

  p1 = cgetg(cl + 1, t_VEC);
  for (j = 1; j <= cl; j++)
  {
    test = 0;
    for (k = 1; (k < j) && !test; k++)
      if (gegal((GEN)cond[j], (GEN)cond[k])) test = k;

    if (test)
      p1[j] = p1[test];
    else
      p1[j] = (long)veceint1(stoi(nn[j]), (GEN)c2[j], prec);

    av2 = avma;
    for (n = 1; n <= nn[j]; n++)
    {
      an = EvalCoeff(gmael3(dtcr, j, 5, 2), matan[j][n], mx);

      gaffect(gadd((GEN)veclprime2[j], gmul(gmael(p1, j, n), an)), (GEN)veclprime2[j]);
    }

    avma = av2;
  }

  if (DEBUGLEVEL) msgtimer("Compute V2");

  FreeArray(matan, cl, nmax);
  avma = av1;

  rep[2] = (long)veclprime2;

  return gerepileupto(av, rep);
}

/* recover a quadratic integer by an exhaustive search */
static GEN
recbeta2(GEN nf,  GEN beta,  GEN bound,  long prec)
{
  long av = avma, av2, tetpil, i, range, G, e, m;
  GEN om, om1, om2, dom, p1, a, b, rom, bom2, *gptr[2];

  G = min( - 20, - bit_accuracy(prec) >> 4);

  if (DEBUGLEVEL > 3)
    fprintferr("\n Precision needed: %ld", G);

  om  = gmael(nf, 7, 2);
  rom = (GEN)nf[6];
  om1 = poleval(om, (GEN)rom[1]);
  om2 = poleval(om, (GEN)rom[2]);
  dom = subrr(om1, om2);

  /* b will run from b to b + range */
  p1 = gaddgs(gmul2n(gceil(absr(divir(bound, dom))), 1), 2);
  range = VERYBIGINT;
  if (cmpis(p1,  VERYBIGINT) < 0)
    range = itos(p1);

  av2 = avma;

  b = gdiv(gsub(bound, beta), dom);
  if (gsigne(b) < 0)
    b = subis(negi(gcvtoi(gneg_i(b), &e)), 1);
  else
    b=gcvtoi(b, &e);

  if (e > 0)  /* precision is lost in truncation */
  {
    avma = av;
    return NULL;
  }

  bom2 = mulir(b, om2);
  m = 0;

  for (i = 0; i <= range; i++)
  {
    /* for every b,  we construct a and test it */
    a = grndtoi(gsub(beta, bom2), &e);

    if (e > 0) /* precision is lost in truncation */
    {
      avma = av;
      return NULL;
    }

    p1 = gsub(mpadd(a, bom2),  beta);

    if ((DEBUGLEVEL > 3) && (expo(p1)<m))
    {
      m = expo(p1);
      fprintferr("\n Precision found: %ld", expo(p1));
    }

    if (gcmp0(p1) || (expo(p1) < G))  /* result found */
    {
      p1 = gadd(a, gmul(b, om));
      return gerepileupto(av, gmodulcp(p1, (GEN)nf[1]));
    }

    tetpil = avma;

    b    = gaddgs(b, 1);
    bom2 = gadd(bom2, om2);

    gptr[0] = &b;
    gptr[1] = &bom2;
    gerepilemanysp(av2, tetpil, gptr, 2);
  }

  /* if it fails... */
  return NULL;
}

/* let polrel define Hk/k,  find L/Q such that Hk=Lk and L and k are
   disjoint */
static GEN
makescind(GEN nf, GEN polabs, long cl, long prec)
{
  long av = avma, i, debkeep = DEBUGLEVEL, fl = 0;
  GEN pol, p1, nf2, dabs, dk;

  if (DEBUGLEVEL == 1) DEBUGLEVEL = 0;

  /* check the result (a little): signature and discriminant */
  dabs = discf(polabs);
  dk   = (GEN)nf[3];

  if (!gegal(dabs, gpuigs(dk, cl)) || (sturmpart(polabs, NULL, NULL)!= 2 * cl))
    err(bugparier, "quadhilbert");

  /* attempt to find the subfields using polred */
  p1 = polred(polabs, (prec << 1) - 2);

  for (i = 1; (i < lg(p1)) && !fl; i++)
  {
    pol = (GEN)p1[i];

    if (degree(pol) == cl)
    {
      if (cl % 2) fl = 1;
      if (!gegal(sqri(discf(pol)), dabs)) fl = 1;
    }
  }
  if (debkeep) msgtimer("polred");

  /* ... if it fails, then use nfsubfields */
  if (!fl)
  {
    nf2 = nfinit0(polabs, 1, prec);
    p1  = subfields(nf2, stoi(cl));

    if (debkeep) msgtimer("subfields");

    for (i = 1; (i < lg(p1)) && !fl; i++)
    {
      pol = gmael(p1, i, 1);

      if (cl % 2) fl = 1;
      if (!gegal(sqri(discf(pol)), (GEN)nf2[3])) fl = 1;
    }

    for (i = 1; (i < lg(p1)) && !fl; i++)
    {
      pol = gmael(p1, i, 1);
      if (degree(gcoeff(nffactor(nf, pol), 1, 1)) == cl) fl = 1;
    }
  }

  if (fl)
  {
    pol = polredabs(pol, prec);
    DEBUGLEVEL = debkeep;

    return gerepileupto(av, pol);
  }

  err(talker, "bug in makescind: no polynomial found");
  return NULL; /* not reached */
}

/* compute the Hilbert class field using genus class field theory when
   the exponent of the class group is 2 */
static GEN
GenusField(GEN bnf, long prec)
{
  long hk, c, l, av = avma;
  GEN disc, quat, x2, pol, div, d;

  hk   = itos(gmael3(bnf, 8, 1, 1));
  disc = gmael(bnf, 7, 3);
  quat = stoi(4);
  x2   = gsqr(polx[0]);

  if (gcmp0(modii(disc, quat))) disc = divii(disc, quat);

  div = divisors(disc);
  c = 1;
  l = 0;

  while(l < hk)
  {
    c++;
    d = (GEN)div[c];

    if (gcmp1(modii(d, quat)))
    {
      if (!l)
	pol = gsub(x2, d);
      else
	pol=(GEN)compositum(pol, gsub(x2, d))[1];

      l = degree(pol);
    }
  }

  return gerepileupto(av, polredabs(pol, prec));
}

/* if flag = 0 returns the reduced polynomial,  flag = 1 returns the
   non-reduced polynomial,  flag = 2 returns an absolute reduced
   polynomial,  flag = 3 returns the polynomial of the Stark's unit,
   flag = -1 computes a fast and crude approximation of the result */
static GEN
AllStark(GEN data,  GEN nf,  long flag,  long newprec)
{
  long cl, i, j, cpt = 0, av, av2, N, h, v, n, dbg, mx;
  int ***matan;
  GEN p1, S, T, polrelnum, polrel, beta, Lp, W, A, veczeta, sig, valchi;
  GEN C, Cmax, dataCR, cond1, L1, *gptr[2], an, pi;


  N     = degree((GEN)nf[1]);
  cond1 = gmael4(data, 1, 2, 1, 2);
  pi    = mppi(newprec);

  v = 1;
  while(gcmp1((GEN)cond1[v])) v++;

LABDOUB:

  av = avma;

  dataCR = (GEN)data[4];
  cl = lg(dataCR)-1;
  mx = GetMaxDeg(dataCR);
  h  = itos(gmul2n(det((GEN)data[2]), -1));

  if (flag >= 0)
  {
    /* compute S,T differently if nf is quadratic */
    if (N == 2)
      p1 = QuadGetST(data, newprec);
    else
      p1 = tmpGetST(dataCR, newprec);

    S = (GEN)p1[1];
    T = (GEN)p1[2];

    Lp = cgetg(cl + 1, t_VEC);
    for (i = 1; i <= cl; i++)
    {
      W = ComputeArtinNumber((GEN)dataCR[i], 0, newprec);
      A = ComputeAChi((GEN)dataCR[i]);

      Lp[i] = lmul(A, gadd(gmul(W, (GEN)S[i]), (GEN)T[i]));
    }

    if (DEBUGLEVEL) msgtimer("Compute W");
  }
  else
  {
    /* we compute only a crude approximation of the result */
    dbg = DEBUGLEVEL;
    DEBUGLEVEL = 0;

    C = cgetg(cl + 1, t_VEC);
    for (i = 1; i <= cl; i++) C[i] = mael(dataCR, i, 2);
    Cmax = vecmax(C);

    n = min(1000, tmpGetBoundN0(Cmax, N, DEFAULTPREC));
    if (dbg) fprintferr("nmax in QuickPol: %ld \n", n);

    /* compute the coefficients an differently if nf is quadratic */
    if (N == 2)
      matan = computean(dataCR, n);
    else
      matan = ComputeCoeff(dataCR, n);

    p1 = cgetg(3, t_COMPLEX);
    p1[1] = lgetr(DEFAULTPREC);
    p1[2] = lgetr(DEFAULTPREC);
    gaffect(gzero, p1);

    L1 = cgetg(cl+1, t_VEC);
    for (i = 1; i <= cl; i++) L1[i] = lcopy(p1);

    /* we use the formulae L(1) = sum (an / n) */
    for (i = 1; i <= cl; i++)
      for (j = 1; j <= n; j++)
      {
	av2 = avma;
	an = EvalCoeff(gmael3(dataCR, i, 5, 2), matan[i][j], mx);
      	gaffect(gadd((GEN)L1[i], gdivgs(an, j)), (GEN)L1[i]);
	avma = av2;
      }

    p1 = gmul2n(gpuigs(mpsqrt(pi), N - 2), 1);

    Lp = cgetg(cl+1, t_VEC);
    for (i = 1; i <= cl; i++)
    {
      W = ComputeArtinNumber((GEN)dataCR[i], 1, DEFAULTPREC);
      A = ComputeAChi((GEN)dataCR[i]);
      W = gmul((GEN)C[i], gmul(A, W));

      Lp[i] = ldiv(gmul(W, gconj((GEN)L1[i])), p1);
    }
  }

  p1 = ComputeLift(gmael(data, 5, 2));

  veczeta = cgetg(h + 1, t_VEC);
  for (i = 1; i <= h; i++)
  {
    veczeta[i] = zero;

    sig = (GEN)p1[i];
    valchi = chiideal(dataCR, sig, 0);

    for (j = 1; j <= cl; j++)
    {
      /* if the character is real or not */
      if (gegal(gdeux, gmael3(dataCR, j, 5, 3)))
	veczeta[i] = ladd((GEN)veczeta[i], gmul((GEN)Lp[j], (GEN)valchi[j]));
      else
	veczeta[i] = ladd((GEN)veczeta[i], gmul2n(greal(gmul((GEN)Lp[j],
							     (GEN)valchi[j])), 1));
    }
    veczeta[i] = ldivgs((GEN)veczeta[i], 2 * h);
  }
  veczeta = greal(veczeta);

  if (DEBUGLEVEL >= 2)
    fprintferr("zetavalues = %Z\n", (long)veczeta);

  polrelnum = gun;
  if (flag < 3)
  {
    /* we want the polynomial of alpha */
    for (j = 1; j <= h; j++)
    {
      p1 = gexp(gmul2n((GEN)veczeta[j], 1), newprec);
      polrelnum = gmul(polrelnum, gsub(polx[0], gadd(p1, ginv(p1))));
    }
  }
  else
  {
    /* we want the polynomial of eps */
    for (j = 1; j <= h; j++)
    {
      p1 = gexp(gmul2n((GEN)veczeta[j], 1), newprec);
      polrelnum = gmul(polrelnum, gsub(polx[0], p1));
      polrelnum = gmul(polrelnum, gsub(polx[0], ginv(p1)));
    }
    /* and its degree is 2 * h */
    h *= 2;
  }

  if (DEBUGLEVEL >= 2)
    fprintferr("polrelnum = %Z\n", (long)polrelnum);
  if (DEBUGLEVEL) msgtimer("Compute polnum");

  if (flag < 0)
  {
    DEBUGLEVEL = dbg;

    if (DEBUGLEVEL >= 2)
      fprintferr("quickpol = %Z\n", (long)polrelnum);
    if (DEBUGLEVEL) msgtimer("QuickPol");

    return gerepileupto(av, gcopy(polrelnum));
  }

  /* we try to recognize this polynomial */
  polrel = RecCoeff(nf, polrelnum, v, flag, newprec);

  if (!polrel) /* if it fails... */
  {
    cpt++;
    if (cpt >= 3) err(talker, "insufficient precision in AllStark");

    /* we compute the precision that we need */
    beta = vecmax(gabs(gtovec(polrelnum), newprec));
    p1   = gceil(gdiv(mplog(beta), dbltor(2.3026)));
    newprec = max(1.2 * newprec, (long)(itos(p1) * pariK1 + 7));

    if (DEBUGLEVEL) err(warnprec, "AllStark", newprec);

    nf = nfnewprec(nf, newprec);
    data[4] = (long)CharNewPrec((GEN)data[4], nf, newprec);

    gptr[0] = &data;
    gptr[1] = &nf;
    gerepilemany(av, gptr, 2);

    goto LABDOUB;
  }

  if (DEBUGLEVEL >= 2)
    fprintferr("polrel = %Z\n", (long)polrel);
  if (DEBUGLEVEL) msgtimer("Recpolnum");

  /* we want a reduced relative polynomial */
  if (!flag) return gerepileupto(av, rnfpolredabs(nf, polrel, 0, newprec));

  /* we just want the polynomial computed */
  if (flag!=2) return gerepileupto(av, gcopy(polrel));

  /* we want a reduced absolute polynomial */
  return gerepileupto(av, rnfpolredabs(nf, polrel, 2, newprec));
}


/********************************************************************/
/*                        Main functions                            */
/********************************************************************/

/* compute the polynomial over Q of the Hilbert class field of
   Q(sqrt(D)) where D is a positive fundamental discriminant */
GEN
quadhilbertreal(GEN D,  long prec)
{
  long av = avma, av1, cl, debkeep = DEBUGLEVEL, newprec;
  GEN pol, bnf, bnrh, p1, nf, exp;

  DEBUGLEVEL = 0;
  timer2();

  /* quick computation of the class number */
  cl = itos((GEN)quadclassunit0(D, 0, NULL, prec)[1]);
  if (cl == 1)
  {
    DEBUGLEVEL = debkeep;

    avma = av;
    return polx[0];
  }

  /* initialize the polynomial defining Q(sqrt{D}) as a polynomial in y */
  pol = quadpoly(D);
  setvarn(pol, fetch_var());

  /* compute the class group */
  bnf = bnfinit0(pol, 1, NULL, prec);
  nf  = (GEN)bnf[7];

  DEBUGLEVEL = debkeep;
  if (DEBUGLEVEL) msgtimer("Compute Cl(k)");

  /* if the exponent of the class group is 2, use rather Genus Field Theory */
  exp = gmael4(bnf, 8, 1, 2, 1);
  if (gegal(exp, gdeux)) { delete_var(); return GenusField(bnf, prec); }

  /* find the modulus defining N */
  bnrh = FindModulus0(bnf, prec);
  if (DEBUGLEVEL) msgtimer("FindModulus");

  av1 = avma;

  if(DEBUGLEVEL >= 2)
  {
    fprintferr("modulus   = %Z\n", mael3(bnrh, 1, 2, 1));
    fprintferr("subgroup  = %Z\n", bnrh[2]);
  }

  /* estimate the precision needed */
  p1 = gsqrt(gmul(gmael(bnf, 7, 3), det(gmael4(bnrh, 1, 2, 1, 1))), prec);
  p1 = gceil(gmul2n(gdiv(p1, dbltor(2.177)), -TWOPOTBITS_IN_LONG));
  newprec = max(itos(p1) + (12 + cl) / BYTES_IN_LONG, prec);

  if (DEBUGLEVEL >= 2) fprintferr("precision = %ld \n", newprec);

  if (newprec > prec)
  {
    nf = nfnewprec(nf, newprec);
    bnrh[4] = (long)CharNewPrec((GEN)bnrh[4], nf, newprec);
  }
  else newprec = prec;

  /* use the generic function AllStark */
  pol = AllStark(bnrh, nf, 2, newprec);
  delete_var();
  return gerepileupto(av, makescind(nf, pol, cl, prec));
}

GEN
bnrstark(GEN bnr,  GEN subgroup,  long flag,  long prec)
{
  long cl, N, newprec, av = avma;
  GEN bnf, dataS, p1, bnrm, D, dataD, dataCR, Mcyc, nf, dataQ, data, polnum, beta;

  bnf  = (GEN)bnr[1];
  nf   = (GEN)bnf[7];
  Mcyc = diagonal(gmael(bnr, 5, 2));
  N    = degree((GEN)nf[1]);

  /* check the bnf */
  if (!varn(gmael(bnf, 7, 1)))
    err(talker, "main variable in bnrstark must not be x");

  if (cmpis(gmael3(bnf, 7, 2, 1), N))
    err(talker, "not a totally real ground base field in bnrstark");

  /* check the bnr */
  checkbnrgen(bnr);

  /* check the subgroup */
  if (gcmp0(subgroup))
    subgroup = Mcyc;
  else
  {
    p1 = gauss(subgroup, Mcyc);
    if (!gcmp1(denom(p1)))
      err(talker, "incorrect subgroup in bnrstark");
  }

  /* TODO: should work even if the modulus is not a conductor */
  if (gcmp0(conductor(bnr, subgroup, -1, prec)))
    err(impl, "bnrstark when the modulus is not the conductor");

  /* check the class field */
  if (!gcmp0(gmael3(bnr, 2, 1, 2)))
    err(talker, "not a totally real class field in bnrstark");

  cl = itos(det(subgroup));
  if (cl == 1) return polx[0];

  timer2();

  /* find a suitable extension N */
  dataS = InitQuotient(bnr, subgroup);
  p1    = FindModulus(dataS, prec);

  bnrm  = (GEN)p1[1];
  D     = (GEN)p1[2];
  dataD = (GEN)p1[3];
  dataQ = (GEN)p1[4];

  if (DEBUGLEVEL) msgtimer("FindModulus");

  if(DEBUGLEVEL >= 2)
  {
    fprintferr("modulus  = %Z\n", mael(bnrm, 2, 1));
    fprintferr("subgroup = %Z\n", (long)D);
  }

  dataCR = InitChar(dataD, prec);

  if (DEBUGLEVEL) msgtimer("InitChar");

  /* prepare the data for AllStark */
  data = cgetg(6, t_VEC);
  data[1] = (long)bnrm;
  data[2] = (long)D;
  data[3] = (long)dataD;
  data[4] = (long)dataCR;
  data[5] = (long)dataQ;


  /* estimate the required precision */
  p1 = gsqrt(gmul(gmael(bnf, 7, 3), det(gmael3(bnrm, 2, 1, 1))), prec);
  p1 = gceil(gmul2n(gdiv(p1, dbltor(2.177)), -TWOPOTBITS_IN_LONG));
  newprec = max(itos(p1) + 3 + (cl >> 2), prec);
  if (DEBUGLEVEL) fprintferr("precision by formulae = %ld \n", newprec);

  if (newprec > 6)
  {
    /* compute a crude approximation of the result to get the required precision */
    polnum = AllStark(data, nf, -1, prec);

    beta = vecmax(gabs(gtovec(polnum), newprec));
    p1 = gceil(gdiv(mplog(beta), dbltor(2.3026)));
    newprec = (long)(itos(p1) * pariK1 + (20 + 2 * cl) / BYTES_IN_LONG);
  }

  if (newprec > prec)
  {
    nf = nfnewprec(nf, newprec);
    mael3(dataD, 1, 1, 7) = (long)nf;

    dataCR  = CharNewPrec(dataCR, nf, newprec);
    data[4] = (long)dataCR;
  }
  else newprec = prec;

  if (DEBUGLEVEL >= 2)
    fprintferr("new precision = %ld \n", newprec);

  return gerepileupto(av, AllStark(data, nf, flag, newprec));
}
